#!/usr/bin/env python
# vim: expandtab sw=4 sts=0:

"""
@copyright: Copyright (C) 2008, 2010 Oracle.  All rights reserved.

@license:   This program is free software; you can redistribute it and/or
            modify it under the terms of the GNU General Public
            License as published by the Free Software Foundation, version 2.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
            General Public License for more details.

@summary:   This is the Oracle Storage Connect Generic NFS plugin.
"""

__docformat__ = 'epytext en'

GENERIC_FS_STAMP = ".generic_fs_stamp"

import re
import os
import stat
import time
import subprocess
import errno
import struct
import platform
import uuid
import fnmatch
from fcntl import ioctl
from gettext import gettext
from cStringIO import StringIO
import logging
from OSCPlugin import *
from os.path import join, exists
import socket 

_ = gettext

__all__ = ['GenericNFSPlugin']

__min_osc_version__ = ["1", "2", "7"]

class GenericNFSPlugin(IFileSystemPlugin):
    """This is the Oracle Generic NFS Plugin"""

    #
    # Required class variables as per the Storage Connect Plugin spec
    #
    plugin_name             = "Oracle Generic Network File System"
    vendor_name             = "Oracle"
    plugin_version          = "1.1.0"
    plugin_desc             = "Oracle Generic Network File System Storage Connect Plugin"
    required_api_vers       = __min_osc_version__
    filesys_type            = IFileSystemPlugin.NetworkFileSystem
    filesys_name            = "NFS"
    plugin_ability          = {"snapshot":              ABILITY_TYPES.UNSUPPORTED,
                               "custom_snap_name":      ABILITY_TYPES.UNSUPPORTED,
                               "snap_is_sync":          ABILITY_TYPES.UNSUPPORTED,
                               "clone":                 ABILITY_TYPES.UNSUPPORTED,
                               "custom_clone_name":     ABILITY_TYPES.UNSUPPORTED,
                               "clone_is_sync":         ABILITY_TYPES.UNSUPPORTED,
                               "resize":                ABILITY_TYPES.UNSUPPORTED,
                               "resize_is_sync":        ABILITY_TYPES.UNSUPPORTED,
                               "splitclone":            ABILITY_TYPES.UNSUPPORTED,
                               "splitclone_is_sync":    ABILITY_TYPES.UNSUPPORTED,
                               "splitclone_while_open": ABILITY_TYPES.UNSUPPORTED,
                               "require_storage_name":  ABILITY_TYPES.NO,
                               "snapclone":             ABILITY_TYPES.UNSUPPORTED,
                               "snapclone_is_sync":     ABILITY_TYPES.UNSUPPORTED,
                               "backing_device_type":   BACKING_DEVICE_TYPES.UNSUPPORTED,
                               "access_control":        ABILITY_TYPES.NO}

    __generic_plugin__      = True

    # Force 4K blocks so that size calculation is easy
    FS_BLOCKSIZE = 4096

    @staticmethod
    def validate(ss_record):
        IPlugin.logger.debug("ENTER: validate()")
        ss_record = _fillSSRecord(**ss_record)
        try:
            _validateSSRecord(ss_record, 'access_host')
        except (MissingKeyEx, InvalidValueEx), e:
            error = _('Invalid SSrecord: %s') % str(e)
            IPlugin.logger.error(error)
            raise InvalidStorageArrayEx(error)

        IPlugin.logger.debug("EXIT : validate() - Success")

    @staticmethod
    def getCapabilities(ss_record):
        IPlugin.logger.debug("ENTER: getCapabilities()")
        IPlugin.logger.debug("EXIT : getCapabilities()")
        return GenericNFSPlugin.plugin_ability;

    @staticmethod
    def getStorageNames(ss_record):
        IPlugin.logger.debug("ENTER: getStorageNames()")
        IPlugin.logger.debug("EXIT : getStorageNames()")
        return []

    @staticmethod
    def getAccessGroups(ss_record, fs_record=None):
        return [{'grp_name': 'all', 'grp_entries': ['all']}]

    @staticmethod
    def getQoSList(ss_record, fs_record):
        return {"priority": 0, "value": "default"}

    @staticmethod
    def getStorageServerInfo(ss_record):
        IPlugin.logger.debug("ENTER: getStorageServerInfo()")
        ret_record = _fillSSInfo(ss_record)
        IPlugin.logger.debug("EXIT : getStorageServerInfo()")
        return ret_record

    @staticmethod
    def getFileSystemInfo(ss_record, fs_record):
        IPlugin.logger.debug("ENTER: getFileSysytemInfo()")
        ret_record = _fillFSInfo(ss_record, fs_record)
        IPlugin.logger.debug("EXIT : getFileSysytemInfo()")
        return ret_record

    @staticmethod
    def getStatus(ss_record, fs_record=None):
        if fs_record:
            status = _getFSStatus(ss_record, fs_record)
        else:
            status = _getModuleStatus()
        return status

    @staticmethod
    def listFileSystems(ss_record, include_unpresented=False):
        #
        # List all mounted files systems and stamp each one with a unique uuid if not
        # already stamped
        #

        IPlugin.logger.debug("ENTER: listFileSysytems()")

        mounted_fs = []
        fs_record_list = []
        
        # List all exported file systems 
       
        exports = _getExports(ss_record)

         # get all mounts

        mounts = _getMounts()

        for mountpoint in mounts:
            mlist = mounts[mountpoint]      
            if mlist[0]['server'] != ss_record['access_host']:
                continue

            # We are looking just for the root export being mounted
            # not the subdir. If it was the subdir that was mounted
            # the fs will be picked up in the unmounted section.

            if mlist[0]['export'] not in exports:
               continue

            fs_record = {}
            fs_record['uuid'] = _getFSStamp(mountpoint)
            stat = os.statvfs(mountpoint)
            fs_record['size'] = (stat.f_bsize * stat.f_blocks)
            fs_record['free_sz'] = (stat.f_bsize * stat.f_bfree)
            fs_record['access_path'] = mlist[0]['device']
            mounted_fs.append(mlist[0]['export'])
            fs_record = _fillFSRecord(ss_record, fs_record)
            fs_record['status'] = 'mounted'
            fs_record['state'] = FS_STATES.MOUNTED
            fs_record['mount_options'] = mlist[0]['options'].split(',')
            fs_record_list.append(fs_record)

        if include_unpresented:

            import tempfile
            path = tempfile.mkdtemp(prefix='/ovs_mount_temp')            
            if os.path.ismount(path):
                p_cmd = ("/bin/umount " + path) 
                p_cmd = p_cmd.split(" ")
                mnt_p = subprocess.Popen(p_cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

                (p_out, p_err) = mnt_p.communicate()

                if mnt_p.returncode != 0:
                    error = _('Unable to umount temp mount point "%s": %s') % (path, p_err)
                    IPlugin.logger.error(error)
                    raise OperationFailedEx(error)

            for export in exports:
                
                if export in mounted_fs:
                    continue

                filesyspath = "%s:%s" % (ss_record['access_host'], export)
                p_cmd = ("/bin/mount "
                         "-t nfs "
                         "-o nosharecache,udp,vers=3,soft,retry=0 "
                         "%s " + path) % filesyspath

                p_cmd = p_cmd.split(" ")
                mnt_p = subprocess.Popen(p_cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

                (p_out, p_err) = mnt_p.communicate()

                if mnt_p.returncode != 0:
                    error = _('Unable to mount "%s": %s') % (filesyspath, p_err)
                    IPlugin.logger.error(error)
                    fs_record = {}
                    fs_record['uuid'] = 'ffffffff-ffff-ffff-ffff-fffffffffff'
                    fs_record_list.append(fs_record)
                    continue

                fs_record = {}
                fs_record['access_path'] = filesyspath
                fs_record['uuid'] = _getFSStamp(path)
                stat = os.statvfs(path)
                fs_record['size'] = (stat.f_bsize * stat.f_blocks)
                fs_record['free_sz'] = (stat.f_bsize * stat.f_bfree)
                fs_record = _fillFSRecord(ss_record, fs_record)
                fs_record['status'] = 'not mounted'
                fs_record['state'] = FS_STATES.UNMOUNTED
                mount = _getMountInfo(filesyspath)
                fs_record['mount_options'] = mount['options'].split(',')
                fs_record_list.append(fs_record)
               
                p_cmd = ("/bin/umount " + path) 

                p_cmd = p_cmd.split(" ")
                mnt_p = subprocess.Popen(p_cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

                (p_out, p_err) = mnt_p.communicate()

                if mnt_p.returncode != 0:
                    error = _('Unable to umount file system "%s": %s') % (fs_record['access_path'], p_err)
                    IPlugin.logger.error(error)
                    raise OperationFailedEx(error)

            os.removedirs(path)

        IPlugin.logger.debug("EXIT : listFileSysytems()")
        return fs_record_list


    @staticmethod
    def listMountPoints(ss_record, fs_record=None):
        
        if fs_record:
            _validateFSRecord(ss_record, fs_record)

        mount_records = []
       
       
        # List all exported file systems 
       
        exports = _getExports(ss_record)

         # get all mounts

        mounts = _getMounts()


        for mountpoint in mounts:
            mlist = mounts[mountpoint]     

            addr1 = socket.gethostbyname(mlist[0]['server'])
            addr2 = socket.gethostbyname(ss_record['access_host'])

            # if mlist[0]['server'] != ss_record['access_host']:
            if addr1 != addr2:
                continue

            mount_uuid = _getFSStampforMount(mountpoint)

            # Look for the root export in device. The root is can be a subset
            # of the mount device. Anything left is subdir.

            sharepath = ''

            for export in exports:
                mounted_export = mlist[0]['export']
                if mounted_export.startswith(export):
                    path = mounted_export.replace(export, "")
                    if path.startswith(os.path.sep):
                        sharepath = path
                        break

            if fs_record and mount_uuid != fs_record['uuid']:
                continue  

            mr = {}
            if len(sharepath):
                mr['share_path'] = sharepath
            mr['fs_uuid'] = mount_uuid
            mr['mount_point'] = mountpoint
            mr['status'] = 'mounted'
            mr['options'] =  mlist[0]['options'].split(',')
        
            mount_records.append(mr)

        return mount_records

    @staticmethod
    def mount(ss_record, fs_record, mountpoint, sharepath=None,  create_mount_point = False, options=[]):

        _validateFSRecord(ss_record, fs_record)
        if not sharepath:
            sharepath=''

        if not mountpoint:
            error = _('Cannot mount filesystem "%s" without a mountpoint') % \
                fs_record['access_path']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if create_mount_point and not exists(mountpoint):
                os.makedirs(mountpoint)
    
        else:
            mounts = GenericNFSPlugin.listMountPoints(ss_record, fs_record)
            for mount in mounts:
                if os.path.samefile(mount['mount_point'], mountpoint):
                    _setFSStampforMount(mountpoint, fs_record['uuid'])
                    return

        full_access_path = fs_record['access_path'] + sharepath

        cmd = ['mount']
        cmd.extend(['-t', 'nfs'])
        if len(options):
            option_string = ','.join(options)
            if len(option_string):
                cmd.extend(['-o', option_string])
        cmd.extend([full_access_path, mountpoint])
 
        IPlugin.logger.debug("cmd = %s" % cmd) 
        mnt_p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        (p_out, p_err) = mnt_p.communicate()
        if mnt_p.returncode != 0:
            error = _('Unable to mount file system "%s": %s') % (full_access_path,
                                                                 p_err)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)

        _setFSStampforMount(mountpoint, fs_record['uuid'])
        fs_record['mount_options'] = options
        return fs_record


    @staticmethod
    def unmount(ss_record, fs_record, mountpoint, destroy_mount_point = False):

        _validateFSRecord(ss_record, fs_record)
        mounts = GenericNFSPlugin.listMountPoints(ss_record, fs_record)
        for mount in mounts:
            if os.path.samefile(mount['mount_point'], mountpoint):
                IPlugin.logger.debug("To umount %s" % mount)
                break
        else:
            IPlugin.logger.debug("%s not mounted, return" % mountpoint)
            return

        umnt_p = subprocess.Popen(["umount", mountpoint],
                                  stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE)
        (p_out, p_err) = umnt_p.communicate()
        if umnt_p.returncode != 0:
            error = _('Unable to unmount filesystem "%s": %s') % (mountpoint,
                                                                  p_err)
            IPlugin.logger.error(error)
            raise OperationFailedEx(error)

        if destroy_mount_point:
            try:
                os.rmdir(mountpoint)
            except:
                pass
        
        IPlugin.logger.debug("Umount completed")

    @staticmethod
    def create(ss_record, fs_record, file_name, file_type, size, sparse=True, force=False):
        _validateFSRecord(ss_record, fs_record)
        if not fs_record['state'] == FS_STATES.MOUNTED:
            error = _('File system "%s" not mounted') % fs_record['name']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        
        if file_type == IFileSystemPlugin.FileType:
            create_op = os.mknod
        elif file_type == IFileSystemPlugin.DirType:
            create_op = os.mkdir
        else:
            error = _('Cannot create file of type "%s"') % file_record['fr_type']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not file_name.startswith('/'):
            mounts = GenericNFSPlugin.listMountPoints(ss_record, fs_record)
            if not mounts:
                error = _('File system "%s" not mounted') % fs_record['name']
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)
            mount = mounts[0]
            file_path = os.path.join(mount['mount_point'], file_name)
        else:
            file_path = file_name
        file_record = _fillFileRecord(fs_record, file_path=file_path, file_sz=0)
        fr = _doCreate(create_op, fs_record, file_record, force)
        if size:
            fr = GenericNFSPlugin.resize(ss_record, fs_record, fr, size, sparse)
        return fr

    @staticmethod
    def destroy(ss_record, fs_record, file_record):
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        _validateFileRecord(fs_record, file_record, file_path=True,
                            fr_type=True)
        
        if file_record['fr_type'] == IFileSystemPlugin.DirType:
            os.rmdir(file_record['file_path'])
        else:
            os.remove(file_record['file_path'])

    @staticmethod
    def resize(ss_record, fs_record, file_record, new_size, sparse=True):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        _validateFileRecord(fs_record, file_record, file_path=True,
                            fr_type=True)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            error = _('Cannot resize file of type "%s"') % file_record['fr_type']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        _resize(fs_record, file_record, new_size, sparse)
        return GenericNFSPlugin.getFileInfo(ss_record, fs_record, file_record)

    @staticmethod
    def getFileInfo(ss_record, fs_record, file_record, file_comp_record=None):
        
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record, file_path=True)
        if file_comp_record:
            error = _('Shared extent comparison not yet supported')
            IPlugin.logger.error(error)
            raise NoSuchOperationEx(error)
        if not _containsFile(ss_record, fs_record, file_record['file_path']):
            error = _('File "%s" is not contained on filesystem "%s"') % (
                file_record['file_path'], fs_record['name'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        st = os.lstat(file_record['file_path'])
        file_record['fr_type'] = _statToFileType(st)
        file_record['ondisk_sz'] = st.st_blocks * 512
        file_record['file_sz' ] = st.st_size
        return _fillFileRecord(fs_record, **file_record)

    @staticmethod
    def list(ss_record, fs_record, file_record, recursive=False):
        
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.DirType:
            error = _('Not a directory: "%s"') % file_record['file_path']
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)
        list_path = file_record['file_path']
        files = []
        def walk_get_fr(dirname, fname):
            full_path = os.path.join(dirname, fname)
            st = os.lstat(full_path)
            ft = _statToFileType(st)
            oz = st.st_blocks * 512
            fz = st.st_size
            return _fillFileRecord(fs_record, file_path=full_path,
                                   fr_type=ft, ondisk_sz=oz, file_sz=fz)
        if recursive:
            def walk_func(l, dirname, fnames):
                for name in fnames:
                    if file_record.has_key('name_pattern'):
                        if not fnmatch.fnmatch(name, file_record['name_pattern']):
                            continue
                    fr = walk_get_fr(dirname, name)
                    l.append(fr)
        else:
            def walk_func(l, dirname, fnames):
                while len(fnames):
                    name = fnames.pop(0)
                    if file_record.has_key('name_pattern'):
                        if not fnmatch.fnmatch(name, file_record['name_pattern']):
                            continue
                    fr = walk_get_fr(dirname, name)
                    l.append(fr)
        os.path.walk(list_path, walk_func, files)
        return files

    @staticmethod
    def copy(ss_record, fs_record, file_record, dest_file_name=None):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)

        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            error = _('Cannot clone file "%s": Invalid file type "%s"') % \
                (file_record['file_path'], file_record['fr_type'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not _containsFile(ss_record, fs_record,
                             file_record['file_path']):
            error = _('Source file "%s" is not contained on filesystem "%s"') % (
                file_record['file_path'], fs_record['name'])
            IPlugin.logger.error(error)
            raise InvalidValueEx(error)

        if not os.path.exists(file_record['file_path']):
            error = _('Unable to locate file "%s" to clone') % \
                file_record['file_path']
            IPlugin.logger.error(error)
            raise CloneFailedEx(error)

        if dest_file_name:
            if not _containsFile(ss_record, fs_record, dest_file_name):
                error = _('Destination file "%s" is not contained on filesystem "%s"') % (
                    dest_file_name, fs_record['name'])
                IPlugin.logger.error(error)
                raise InvalidValueEx(error)
            if not os.path.exists(os.path.dirname(dest_file_name)):
                os.makedirs(os.path.dirname(dest_file_name))
        else:
            dest_file_name = "%s.%s" % (file_record['file_path'],
                                        time.strftime("%Y%m%d%H%M%S"))

        _fileClone(fs_record, file_record['file_path'], dest_file_name)
        return _fillFileRecord(fs_record, file_path=dest_file_name)

    @staticmethod
    def getFileSystemCloneLimits(ss_record, fs_record = None):
        return -1

    @staticmethod
    def getCloneLimits(ss_record, fs_record, file_record):
        if file_record['fr_type'] == IFileSystemPlugin.FileType:
            return -1
        else:
            return 0

    @staticmethod
    def isCloneable(ss_record, fs_record, file_record):
        _validateSSRecord(ss_record)
        _validateFSRecord(ss_record, fs_record)
        _validateFileRecord(fs_record, file_record)
        if file_record['fr_type'] != IFileSystemPlugin.FileType:
            return False
        if _containsFile(ss_record, fs_record, file_record["file_path"]):
            return True

        return False

    @staticmethod
    def getCurrentClones(ss_record, fs_record, file_record=None):
        return []

    
    #########################################################################
    ###             START OF UNSUPPORTED APIs                             ###
    #########################################################################

    @staticmethod
    def isRestorable(ss_record, fs_record, file_record, snap_file_record):
        IPlugin.logger.debug("ENTER: isRestorable()")
        IPlugin.logger.error("isRestorable() is not supported!")
        raise NoSuchOperationEx(_("isRestorable() is not supported!"))


    @staticmethod
    def createSnap(ss_record, fs_record, file_record, snap_name=None):
        
        IPlugin.logger.debug("ENTER: createSnap()")
        IPlugin.logger.error("createSnap() is not supported!")
        raise NoSuchOperationEx(_("createSnap() is not supported!"))
    

    @staticmethod
    def splitClone(ss_record, fs_record, file_record):

        IPlugin.logger.debug("ENTER: splitClone()")
        IPlugin.logger.error("splitClone() is not supported!")
        raise NoSuchOperationEx(_("splitClone() is not supported!"))
      

    @staticmethod
    def cloneFromSnap(ss_record, fs_record, snap_file_record,
                      dest_file_name=None):
        IPlugin.logger.debug("ENTER: cloneFromSnap()")
        IPlugin.logger.error("cloneFromSnap() is not supported!")
        raise NoSuchOperationEx(_("cloneFromSnap() is not supported!"))
        
   
    @staticmethod
    def getFileSystemSnapLimits(ss_record, fs_record = None):

        IPlugin.logger.debug("ENTER: getFileSystemSnapLimits()")
        IPlugin.logger.error("getFileSystemSnapLimits() is not supported!")
        raise NoSuchOperationEx(_("getFileSystemSnapLimits() is not supported!"))
       

    @staticmethod
    def getSnapLimits(ss_record, fs_record, file_record):

        IPlugin.logger.debug("ENTER: getSnapLimits()")
        IPlugin.logger.error("getSnapLimits() is not supported!")
        raise NoSuchOperationEx(_("getSnapLimits() is not supported!"))
        

    @staticmethod
    def isSnapable(ss_record, fs_record, file_record):
        
        IPlugin.logger.debug("ENTER: isSnapable()")
        IPlugin.logger.error("isSnapable() is not supported!")
        raise NoSuchOperationEx(_("isSnapable() is not supported!"))


    @staticmethod
    def getFileSystemCloneLimits(ss_record, fs_record = None):

        IPlugin.logger.debug("ENTER: getFileSystemCloneLimits()")
        IPlugin.logger.error("getFileSystemCloneLimits() is not supported!")
        raise NoSuchOperationEx(_("getFileSystemCloneLimits() is not supported!"))
        

    @staticmethod
    def snapRestore(ss_record, fs_record, snap_file_record):
        IPlugin.logger.debug("ENTER: snapRestore()")
        IPlugin.logger.error("snapRestore() is not supported!")
        raise NoSuchOperationEx(_("snapRestore() is not supported!"))


    @staticmethod
    def snapRemove(ss_record, fs_record, snap_file_record):
        IPlugin.logger.debug("ENTER: snapRemove()")
        IPlugin.logger.error("snapRemove() is not supported!")
        raise NoSuchOperationEx(_("snapRemove() is not supported!"))


    @staticmethod
    def getCurrentSnaps(ss_record, fs_record, file_record=None):
        IPlugin.logger.debug("ENTER: getCurrentSnaps()")
        IPlugin.logger.error("getCurrentSnaps() is not supported!")
        raise NoSuchOperationEx(_("getCurrentSnaps() is not supported!"))


    @staticmethod
    def createFileSystem(ss_record, name, backing_device, size=0,
                         access_grp_names=None, se_record_ids=None,
                         qos=None, force=False):
        IPlugin.logger.debug("ENTER: createFileSystem()")
        IPlugin.logger.error("createFileSystem() is not supported!")
        raise NoSuchOperationEx(_("createFileSystem() is not supported!"))

        
    @staticmethod
    def resizeFileSystem(ss_record, fs_record, new_size):
        IPlugin.logger.debug("ENTER: createFileSystem()")
        IPlugin.logger.error("createFileSystem() is not supported!")
        raise NoSuchOperationEx(_("createFileSystem() is not supported!"))


    @staticmethod
    def destroyFileSystem(ss_record, fs_record):
        IPlugin.logger.debug("ENTER: destroyFileSystem()")
        IPlugin.logger.error("destroyFileSystem() is not supported!")
        raise NoSuchOperationEx(_("destroyFileSystem() is not supported!"))

      


    #########################################################################
    ###             START OF THE PLUGIN PRIVATE METHODS                   ###
    #########################################################################


def _validateRecord(name, record, key_hash):
    '''Validate a record.  Any key in key_hash must exist in the record.
       if key_hash[key] is True, the record must have a non-False value
       for it.'''
    for key in key_hash:
        if not record.has_key(key):
            raise MissingKeyEx(_('%s does not have key "%s"') % (name, key),
                               key)
        elif key_hash[key] and not record[key]:
            raise MissingKeyEx(_('%s does not have a value for key "%s"') %
                               (name, key), key)

def _buildKeyHash(check_keys, require_value_keys, record_fields):
    '''Build a key_hash for _validateRecord().  record_fields is only used
       if there are no keys in the other lists.  It is the record
       definition, and if record_fields[key][1] is true then key is treated
       like it exists in check_keys.'''
    key_hash = {}
    for key in check_keys:
        key_hash[key] = False
    for key in require_value_keys.iterkeys():
        key_hash[key] = require_value_keys[key]
    if not len(key_hash.keys()):
        for (key, value) in record_fields.iteritems():
            if value[1]:
                key_hash[key] = False
    return key_hash

def _validateSSRecord(ss_record, *check_keys, **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__SSRecord_fields__)
    _validateRecord('SSRecord', ss_record, key_hash)
    if ss_record["storage_type"] != IFileSystemPlugin.FileSysStorage:
        raise InvalidValueEx(_('Invalid filesystem type: "%s"') %
                             ss_record['storage_type'])
    

def _validateFSRecord(ss_record, fs_record, *check_keys,
                      **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__FSRecord_fields__)
    _validateRecord('FSRecord', fs_record, key_hash)
    if key_hash.has_key('filesys_type') and \
        fs_record['filesys_type'] != IFileSystemPlugin.NetworkFileSystem:
        raise InvalidValueEx(_('Invalid filesystem type: "%s"') %
                             fs_record['filesys_type'])
    if key_hash.has_key('ss_uuid') and \
        fs_record['ss_uuid'] != ss_record['uuid']:
        raise InvalidValueEx(_('Invalid Storage Server uuid: expected "%s", got "%s"') %
                             (ss_record['uuid'], fs_record['ss_uuid']))

def _validateMountRecord(fs_record, mount_record, *check_keys,
                         **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__MountRecord_fields__)
    _validateRecord('MountRecord', mount_record, key_hash)
    if fs_record and key_hash.has_key('fs_uuid') and \
        mount_record['fs_uuid'] != fs_record['uuid']:
        raise InvalidValueEx(_('Invalid Filesystem uuid: expected "%s", got "%s"') %
                             (fs_record['uuid'], mount_record['ss_uuid']))

def _validateFileRecord(fs_record, file_record, *check_keys,
                        **require_value_keys):
    key_hash = _buildKeyHash(check_keys, require_value_keys,
                             IFileSystemPlugin.__FileRecord_fields__)
    _validateRecord('FileRecord', file_record, key_hash)
    if key_hash.has_key('fs_uuid') and \
        file_record['fs_uuid'] != fs_record['uuid']:
        raise InvalidValueEx(_('Invalid Filesystem uuid: expected "%s", got "%s"') %
                             (fs_record['uuid'], file_record['ss_uuid']))
    if key_hash.has_key('file_path') and \
        file_record['file_path'][0] != '/':
        raise InvalidValueEx(_('Path "%s" is not absolute') %
                             file_record['file_path'])


def _statToFileType(st):
    if stat.S_ISDIR(st.st_mode):
        return IFileSystemPlugin.DirType
    elif stat.S_ISBLK(st.st_mode):
        return IFileSystemPlugin.BlkDevType
    elif stat.S_ISCHR(st.st_mode):
        return IFileSystemPlugin.CharDevType
    elif stat.S_ISLNK(st.st_mode):
        return IFileSystemPlugin.SymLinkType
    elif stat.S_ISSOCK(st.st_mode):
        return IFileSystemPlugin.SockFileType
    elif stat.S_ISFIFO(st.st_mode):
        return IFileSystemPlugin.FifoType
    else:
        return IFileSystemPlugin.FileType

def _procMountsToMount(line):
    l = line.split()
    return {
        'device': l[0].decode('string_escape'),
        'mountpoint': l[1].decode('string_escape'),
        'fstype': l[2],
        'options': l[3].decode('string_escape')
    }


def _getFSStamp(mount_point):
    fs_uuid = 'ffffffff-ffff-ffff-ffff-fffffffffff'

    stamp_file = join(mount_point, GENERIC_FS_STAMP)
    if not exists(stamp_file):
        try:
            file = open(stamp_file, "w")
        except:
            error = _('Warning cannot create .generic_fs_stamp on %s') % mount_point
            IPlugin.logger.error(error)
            return fs_uuid
        fs_uuid = str(uuid.uuid4())
        file.write(fs_uuid)
        file.close
    else:
        try:
            file = open(stamp_file, "r")
        except:
            error = _('Warning cannot open .generic_fs_stamp on %s') % mount_point
            IPlugin.logger.error(error)
            return fs_uuid
        fs_uuid = file.read().strip()
        file.close
        if not fs_uuid:
            try:
                file = open(stamp_file, "w")
            except:
                error = _('Warning cannot open .generic_fs_stamp on %s') % mount_point
                IPlugin.logger.error(error)
                return fs_uuid
            fs_uuid = str(uuid.uuid4())
            file.write(fs_uuid)
            file.close

    return fs_uuid


def _getFSStampforMount(mount_point):
    no_fs_uuid = 'ffffffff-ffff-ffff-ffff-fffffffffff'

    stamp_file = join(mount_point, GENERIC_FS_STAMP)
    if not exists(stamp_file):
        error = _('Warning .generic_fs_stamp on %s does not exist') % mount_point
        IPlugin.logger.error(error)
        return no_fs_uuid
    else:
        try:
            file = open(stamp_file, "r")
        except:
            error = _('Warning Unable to open .generic_fs_stamp on %s') % mount_point
            IPlugin.logger.error(error)
            return no_fs_uuid
        file_fs_uuid = file.read().strip()
        file.close
        if not file_fs_uuid:
            return no_fs_uuid
            

    return file_fs_uuid

def _setFSStampforMount(mount_point, fs_uuid):
    no_fs_uuid = 'ffffffff-ffff-ffff-ffff-fffffffffff'

    stamp_file = join(mount_point, GENERIC_FS_STAMP)
    if not exists(stamp_file):
        try:
            file = open(stamp_file, "w")
        except:
            error = _('Warning cannot create .generic_fs_stamp on %s') % mount_point
            IPlugin.logger.error(error)
            return no_fs_uuid
       
        file.write(fs_uuid)
        file.close
    else:
        try:
            file = open(stamp_file, "r")
        except:
            error = _('Warning cannot open .generic_fs_stamp on %s') % mount_point
            IPlugin.logger.error(error)
            return no_fs_uuid
        file_fs_uuid = file.read().strip()
        file.close
        if not file_fs_uuid:
            try:
                file = open(stamp_file, "w")
            except:
                error = _('Warning cannot open .generic_fs_stamp on %s') % mount_point
                IPlugin.logger.error(error)
                return no_fs_uuid
            
            file.write(fs_uuid)
            file.close
        else:
            fs_uuid = file_fs_uuid

    return fs_uuid

def _getMounts():
    mounts = {}
    try:
        try:
            proc_mounts = open("/proc/mounts", "r")
            for line in proc_mounts:
                mount = _procMountsToMount(line)
                if mount['fstype'] != 'nfs':
                    continue
                server, export = mount['device'].split(":")
                mount['server'] = server
                mount['export'] = export
                if mounts.has_key(mount['mountpoint']):
                    mounts[mount['mountpoint']].append(mount)
                else:
                    mounts[mount['mountpoint']] = [mount]
                
        except IOError, ioe:
            proc_mounts.close()
            raise
    finally:
        proc_mounts.close()
    return mounts

def _getMountInfo(device):
   
    try:
        try:
            proc_mounts = open("/proc/mounts", "r")
            for line in proc_mounts:
                mount = _procMountsToMount(line)
                if mount['fstype'] != 'nfs':
                    continue
                if mount['device'] != device:
                    continue
                server, export = mount['device'].split(":")
                mount['server'] = server
                mount['export'] = export
                
                
        except IOError, ioe:
            proc_mounts.close()
            raise
    finally:
        proc_mounts.close()
    return mount

def _getExports(ss_record):

    exports = []

    p_cmd = ("/usr/sbin/showmount "
             "--exports "
             "%s") % ss_record['access_host']

    p_cmd = p_cmd.split(" ")
    sh_p = subprocess.Popen(p_cmd,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE)

    (p_out, p_err) = sh_p.communicate()

    if sh_p.returncode != 0:
        IPlugin.logger.error("Unable to access file system at %s: %s" %
                             (ss_record['access_host'], p_err))
        raise OperationFailedEx(_("Unable to access file system at %s: %s" %
                             (ss_record['access_host'], p_err)))

    export_items = p_out.splitlines()

    export_items = export_items[1:]

    for export in export_items:
        exports.append(export.split()[0])

    return exports


def _getFSStatus(ss_record, fs_record):
    
    _validateFSRecord(ss_record, fs_record)
    if len(GenericNFSPlugin.listMountPoints(ss_record, fs_record)):
        fs_record['status'] = 'mounted'
    else:
        fs_record['status'] = 'not mounted'
        
    return fs_record['status']

def _fillFSInfo(ss_record, fs_record):
    _validateSSRecord(ss_record, 'access_host')
    _validateFSRecord(ss_record, fs_record, "access_path", "ss_uuid")
 
    if not fs_record.get('name', ''):
        fs_record['name'] = fs_record['uuid']
    fs_record['status'] = 'unknown'
    fs_record['state'] = FS_STATES.UNKNOWN
    fs_record["access_grp_names"] = "all"
    mount = _getMountInfo(['access_path'])
    if len(mount):
        fs_record['status'] = 'mounted'
        fs_record['state'] = FS_STATES.MOUNTED
        fs_record['mount_options'] = mount['options'].split(',')
    else:
        fs_record['status'] = 'not mounted'
        fs_record['state'] = FS_STATES.UNMOUNTED
    return fs_record

def _getModuleStatus():
    f = file("/proc/filesystems", "r")
    try:
        for l in f:
            if l.find('nfs') >= 0:
                return "loaded"
        return "not loaded"
    finally:
        f.close()

def _fillSSInfo(ss_record):
    ss_record['total_sz'] = -1
    ss_record['used_sz'] = -1
    ss_record['free_sz'] = -1
    ss_record['access_grps'] = []
    ss_record['storage_desc'] = "The NFS File system"
    ss_record['status'] = _getModuleStatus()
    return ss_record

def _fillSSRecord(**kwargs):
    '''Take attributes for an ss_record and fill it in.  This makes sure
       everything needed was specified.'''
    ss_record = kwargs
    if not ss_record.has_key('access_grps'):
        ss_record['access_grps'] = []
    if not ss_record.has_key('storage_desc'):
        ss_record['storage_desc'] = 'The NFS File system'
    for key in ['admin_host', 'admin_user', 'admin_passwd', 'total_sz',
                'used_sz', 'free_sz', 'status']:
        if not ss_record.has_key(key):
            ss_record[key] = IFileSystemPlugin.__SSRecord__[key]
    _validateSSRecord(ss_record)
    return ss_record

def _fillFSRecord(ss_record, fs_record):
    '''Take attributes for a fs_record and fill it in.  This makes sure
       everything needed was specified.'''
    fs_record['ss_uuid'] = ss_record['uuid']
    fs_record['filesys_type'] = IFileSystemPlugin.NetworkFileSystem
    fs_record['access_grp_names'] = 'all'
    if not fs_record.get('name', ''):
        fs_record['name'] = fs_record['uuid']
    if not fs_record.has_key('status'):
        fs_record['status'] = 'unknown'
    if not fs_record.has_key('state'):
        fs_record['state'] = FS_STATES.UNKNOWN
    for key in ['size']:
        if not fs_record.has_key(key):
            fs_record[key] = IFileSystemPlugin.__FSRecord__[key]
    return fs_record


def _fillFileRecord(fs_record, **kwargs):
    '''Take attributes for a file record and fill it in.  This makes sure
       everything needed was specified'''
    file_record = kwargs
    file_record['fs_uuid'] = fs_record['uuid']
    if file_record.has_key('snap_name'):
        file_record['fr_type'] = IFileSystemPlugin.SnapType
    elif not file_record.has_key('fr_type'):
        file_record['fr_type'] = IFileSystemPlugin.FileType
    for key in ['ondisk_sz', 'file_sz']:
        if not file_record.has_key(key):
            file_record[key] = IFileSystemPlugin.__FileRecord__[key]
    _validateFileRecord(fs_record, file_record)
    return file_record



def _doCreate(create_op, fs_record, file_record, force):
    try:
        create_op(file_record['file_path'])
        return _fillFileRecord(fs_record, **file_record)
    except OSError, e:
        if e.errno != errno.EEXIST or not force:
            raise
    if stat.S_ISDIR(os.stat(file_record['file_path']).st_mode):
            os.rmdir(file_record['file_path'])
    else:
            os.remove(file_record['file_path'])
    create_op(file_record['file_path'])
    return _fillFileRecord(fs_record, **file_record)


def _bytesToBlocks(in_bytes):
    return ((in_bytes + (GenericNFSPlugin.FS_BLOCKSIZE - 1)) /
            GenericNFSPlugin.FS_BLOCKSIZE)

def _containsFile(ss_record, fs_record, file_path):
    """
    Check if the file is contained on the file system identified by the
    File System record.

    @param fs_record: L{File System record<OSCPlugin.IPlugin.__FSRecord__>}
    @type  fs_record: C{dict}

    @param file_path: File path to check.
    @type  file_path: C{str}

    @return: True if the file is contained in the file system.
    @rtype:  C{bool}

    @raise IPluginException: on failure
    """
    if not file_path.startswith('/'):
        raise InvalidValueEx(_('Path "%s" is not absolute') % file_path)
    mounts = GenericNFSPlugin.listMountPoints(ss_record, fs_record)
    if not len(mounts):
        raise OperationFailedEx(_('Filesystem "%s" is not mounted') %
                                      fs_record['name'])

      
    for mount in mounts:         
        try:
            mount_stat = os.stat(mount['mount_point'])
            file_stat = os.stat(file_path)
            IPlugin.logger.debug("_containsFile mount stat %s mount_point %s" % (mount_stat.st_dev, mount['mount_point']))
            IPlugin.logger.debug("_containsFile file stat %s file path %s" % (file_stat.st_dev, file_path))
            if mount_stat.st_dev == file_stat.st_dev:
                return True
            
        except OSError, e:
            IPlugin.logger.debug("_containsFile exception %s" % e.strerror)
            if e.errno != errno.ENOENT:
                raise
        new_path = os.path.dirname(file_path)
        # Have we just tried '/'?
        IPlugin.logger.debug("_containsFile new_path %s file_path %s" % (new_path, file_path))
        if new_path == file_path:
            break
        file_path = new_path

    IPlugin.logger.debug("_containsFile mount %s" % mount['mount_point']) 
    
    return False

def _resize(fs_record, file_record, new_size, sparse):
    if file_record['file_sz'] == new_size:
        return

    fd = os.open(file_record['file_path'], os.O_WRONLY)
    try:
        if (new_size < file_record['file_sz']) or sparse:
            os.ftruncate(fd, long(new_size))
        else:
            os.lseek(fd, 0, 2)
            left = new_size - file_record['file_sz']
            while left > 0:
                count = pow(2, 20)
                if count > left:
                    count = left
                os.write(fd, '\0' * count)
                left -= count
            os.fsync(fd)
    finally:
        os.close(fd)
    
        
           
        
def _fileClone(fs_record, src_path, dest_path):

    p_cmd = ("/bin/dd "
           "if=%s "
           "of=%s "
           "bs=4096k "
           "oflag=direct") % (src_path, dest_path)
    
    p_cmd = p_cmd.split(" ")
    dd_p = subprocess.Popen(p_cmd,
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE)
    (p_out, p_err) = dd_p.communicate()

    if dd_p.returncode != 0:
        error = _('Unable to clone file "%s" to "%s": %s') % (
            src_path, dest_path, p_err)
        IPlugin.logger.error(error)
        raise CloneFailedEx(error)
