#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2010 Oracle.  All rights reserved.

@license:   This program is free software; you can redistribute it and/or
            modify it under the terms of the GNU General Public
            License as published by the Free Software Foundation, version 2.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
            General Public License for more details.

            You should have received a copy of the GNU General Public
            License along with this program; if not, write to the Free Software
            Foundation, Inc., 59 Temple Place - Suite 330, Boston,
            MA 021110-1307, USA

@summary:   This is the implementation of the Oracle VM Storage Connect
            Plug-in for Generic iSCSI and FibreChannel.
"""

__docformat__ = 'epytext en'

import re
import os
import sys
import time
import socket
import struct
import gettext
import fileinput
import subprocess

import OSCPluginUtils
from cStringIO import StringIO

from OSCPlugin import *

SCSI_GOOD = 0
SCSI_CHECK = 5

vg_GenericISCSI = "Generic_iSCSI_Volume_Group"
vg_GenericSAN = "Generic_SAN_Volume_Group"

#
# Private module variables
#
_ = gettext.gettext
__min_osc_version__ = ["1", "2", "7"]

class GenericPlugin(IStorageArrayPlugin):
    """
    This is the Oracle Generic FC / iSCSI Generic Plugin
    """

    #
    # Required class variables as per the Storage Connect Plugin spec
    #
    plugin_name             = "Oracle Generic SCSI Plugin"
    vendor_name             = "Oracle"
    plugin_version          = "1.1.0"
    plugin_desc             = "Oracle Storage Connect Plugin for Generic FC and iSCSI"
    required_api_vers       = __min_osc_version__
    storage_types           = [IStorageArrayPlugin.SANStorage,
                               IStorageArrayPlugin.iSCSIStorage]
    plugin_ability          = {"snapshot":              ABILITY_TYPES.UNSUPPORTED,
                               "custom_snap_name":      ABILITY_TYPES.UNSUPPORTED,
                               "snap_is_sync":          ABILITY_TYPES.UNSUPPORTED,
                               "clone":                 ABILITY_TYPES.UNSUPPORTED,
                               "custom_clone_name":     ABILITY_TYPES.UNSUPPORTED,
                               "clone_is_sync":         ABILITY_TYPES.UNSUPPORTED,
                               "resize":                ABILITY_TYPES.UNSUPPORTED,
                               "resize_is_sync":        ABILITY_TYPES.UNSUPPORTED,
                               "splitclone":            ABILITY_TYPES.UNSUPPORTED,
                               "splitclone_is_sync":    ABILITY_TYPES.UNSUPPORTED,
                               "snapclone":             ABILITY_TYPES.UNSUPPORTED,
                               "snapclone_is_sync":     ABILITY_TYPES.UNSUPPORTED,
                               "require_storage_name":  ABILITY_TYPES.UNSUPPORTED,
                               "access_control":        ABILITY_TYPES.NO}
    
    __generic_plugin__      = True

    @staticmethod
    def validate(ss_record):
        IPlugin.logger.debug("ENTER: validate()")
        missing_key = ""

        if not ss_record.get("uuid", None):
            missing_key += " uuid"

        if not ss_record.get("name", None):
            missing_key += " name"

        if missing_key:
            IPlugin.logger.error("validate() - '%s' have to be supplied!" %
                                 missing_key)
            raise MissingKeyEx(_("'%s' have to be supplied" % missing_key),
                                 missing_key)

        if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
            IPlugin.logger.debug("IStorageArrayPlugin.iSCSIStorage")
            missing_key = ""
            if not ss_record.get("access_host", None):
                missing_key += " access_host"

            if not ss_record.get("access_port", None):
                missing_key += " access_port"

            if ss_record.get("chap", False):
                if not ss_record.get("username", None):
                    missing_key += " username"

                if not ss_record.get("passwd", None):
                    missing_key += " passwd"

            if missing_key:
                IPlugin.logger.error("validate() - '%s' have to be supplied!" %
                                     missing_key)
                raise MissingKeyEx(_("'%s' have to be supplied" % missing_key),
                                     missing_key)

        elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
            IPlugin.logger.debug("IStorageArrayPlugin.SANStorage")

        else:
            IPlugin.logger.error("validate() - Unknown storage_type %s "
                                 "specified." % ss_record["storage_type"])
            raise InvalidValueEx(_("Unknown storage_type %s specified" %
                                   ss_record["storage_type"]))

        
        IPlugin.logger.debug("EXIT : validate() - Success")


    @staticmethod
    def getCapabilities(ss_record):
        IPlugin.logger.debug("ENTER: getCapabilities()") 
        IPlugin.logger.debug("EXIT : getCapabilities()")
        return GenericPlugin.plugin_ability


    @staticmethod
    def getInfo(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getInfo()")

        if se_record:

            if not se_record.get("uuid", None):
                IPlugin.logger.error("getInfo - No uuid")
                raise InvalidValueEx(_("No uuid"))


            dev_name = "%s/%s" % (IPlugin.dev_path_prefix, se_record["uuid"])

            if not os.path.exists(dev_name):
                IPlugin.logger.error("Device %s does not exist %s", dev_name)
                raise InvalidValueEx(_("Device %s does not exist %s", dev_name))

            try:
                lun_info = GenericPlugin._get_lun_info_(dev_name)

            except IOError, e:
                IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                     "info from %s/%s",
                                     e.strerror,
                                     dev_name)

            se_record["se_type"] = IStorageArrayPlugin.LUNType

            se_record["array_path"] = dev_name

            se_record["id"] = lun_info["serial"]

            if lun_info.has_key("page83_id"):
                se_record["page83_id"] = OSCPluginUtils.getPage83FromMPDev(dev_name)  

            se_record["size"] = lun_info["size"]

            se_record["status"] = GenericPlugin.getStatus(ss_record, se_record)

            se_record["path"] = [dev_name]

            if se_record.get("vol_group_name", None):
                if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                    se_record["vol_group_name"] = vg_GenericISCSI
                else:
                    se_record["vol_group_name"] = vg_GenericSAN

            IPlugin.logger.debug("EXIT : getInfo() - Success")
            return se_record

        else:
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                ss_record["storage_desc"] = "Generic iSCSI Storage"                                       
 
            elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:         
                ss_record["storage_desc"] = "Generic Fibre Channel Storage"
                
            ss_record["status"] = GenericPlugin.getStatus(ss_record)

            groups = []
                           
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                groups.append({
                              "vol_name":      vg_GenericISCSI,
                              "vol_total_sz":  0,
                              "vol_used_sz":   0,
                              "vol_free_sz":   0,
                              "vol_alloc_sz":  0,
                              })
            else:
                groups.append({
                              "vol_name":      vg_GenericSAN,
                              "vol_total_sz":  0,
                              "vol_used_sz":   0,
                              "vol_free_sz":   0,
                              "vol_alloc_sz":  0,
                              })
                
            ss_record["vol_groups"] = groups

            IPlugin.logger.debug("EXIT : getInfo() - Success")
            return ss_record


    @staticmethod 
    def list(ss_record, se_type = IPlugin.LUNType):

        IPlugin.logger.debug("ENTER: list()")
        se_record_list = []

        if not os.path.exists(IPlugin.dev_path_prefix):
            IPlugin.logger.debug("EXIT : list() - Nothing to do!")
            return se_record_list

        for dev_name in os.listdir(IPlugin.dev_path_prefix):
            
            if dev_name == 'control':
                continue

            try:
                lun_info = GenericPlugin._get_lun_info_("%s/%s" % (IPlugin.dev_path_prefix, dev_name))
    
            except IOError, e:
                IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                     "info from %s/%s",
                                     e.strerror,
                                     IPlugin.dev_path_prefix,
                                     dev_name)
                continue
    
            if not GenericPlugin._get_dev_server(ss_record, dev_name):
                continue

            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                vg = vg_GenericISCSI
            else:
                vg = vg_GenericSAN

            se_type = IStorageArrayPlugin.LUNType
            se_record = {"se_type":         se_type,
                         "ss_uuid":         ss_record["uuid"],
                         "uuid":            dev_name,
                         "id":              lun_info["serial"],
                         "path":            ["%s/%s" % (IPlugin.dev_path_prefix,dev_name)],
                         "array_path":      "%s/%s" % (IPlugin.dev_path_prefix,dev_name),
                         "size":            lun_info["size"],
                         "vol_group_name":  vg,}

          
            se_record["page83_id"] = OSCPluginUtils.getPage83FromMPDev("%s/%s" % (IPlugin.dev_path_prefix, dev_name))
                
            se_record["status"] = GenericPlugin.getStatus(ss_record, se_record)
                       
            se_record_list.append(se_record)

        IPlugin.logger.debug("EXIT : list() - Success")
        return se_record_list


    @staticmethod 
    def updateSERecords(ss_record, se_records, append_allowed = False):
        IPlugin.logger.debug("ENTER: updateSERecord()")

        if len(se_records) < 1 :
            IPlugin.logger.debug("EXIT : updateSERecord() - Nothing to do!")
            return se_records

        for se_record in se_records:

            if dev_name == 'control':
                continue

            try:
                se_record = GenericPlugin.getInfo(ss_record, se_record)

            except IOError, e:
                IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                     "info from %s/%s",
                                     e.strerror,
                                     se_record["uuid"])
                continue
            

        IPlugin.logger.debug("EXIT : updateSERecord() - Success")
        return se_records


    @staticmethod
    def getStatus(ss_record, se_record = None):

        IPlugin.logger.debug("ENTER: getStatus()")
        
        status = "no status"

        if not se_record:
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                status = GenericPlugin._get_iscsi_service_status()

            elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
                status = GenericPlugin._get_fc_status()

            else:
                raise InvalidValueEx(_("Unsupported storage type is specified"))
                
                
        else:
            if se_record["se_type"] == IStorageArrayPlugin.LUNType: 

                dev_name = "%s/%s" % (IPlugin.dev_path_prefix, se_record["uuid"])
                 
                # This is the SG_IO TUR CDB
                get_sg_tur = "00 00 00 00 00 00"
                p_cmd = ("/usr/bin/sg_raw "
                         "--binary "
                         "--nosense "
                         "--timeout=15 "
                         "%s " +
                         get_sg_tur) % dev_name
                p_cmd = p_cmd.split(" ")
                sgraw_p = subprocess.Popen(p_cmd,
                                             stdout = subprocess.PIPE,
                                             stderr = subprocess.PIPE)

                (p_out, p_err) = sgraw_p.communicate()

                if p_err.find('Good') >= 0:
                    IPlugin.logger.debug("EXIT : getStatus() - Success")
                    status = 'online'

                if p_err.find('Check') >= 0:
                    IPlugin.logger.debug("EXIT : getStatus() - Success")
                    status = 'offline'

                if sgraw_p.returncode != 0:
                    raise IOError("Unable to issue SG_IO TUR CDB to %s: %s" % (dev_name ,p_err))

            else:
                raise InvalidValueEx(_("Unsupported storage element type is specified"))

        IPlugin.logger.debug("EXIT : getStatus() - Success")
        
        return status
           
    @staticmethod
    def getCloneLimits(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getCloneLimits()")
        IPlugin.logger.debug("EXIT : getCloneLimits()")
        return -1
        
    @staticmethod
    def isCloneable(ss_record, se_record):
        IPlugin.logger.debug("ENTER: isCloneable()")
        if se_record["se_type"] == IStorageArrayPlugin.LUNType:
            IPlugin.logger.debug("EXIT : isCloneable() - True")
            return True

        IPlugin.logger.debug("EXIT : isCloneable() - False")
        return False


    @staticmethod 
    def copy(ss_record, se_record, dest_se_record, qos = None):
        IPlugin.logger.debug("ENTER: copy()")

        p_cmd = ("/bin/dd "
               "if=%s "
               "of=%s "
               "bs=4096k "
               "oflag=direct") % (se_record["path"], dest_se_record["path"])
        
        p_cmd = p_cmd.split(" ")
        dd_p = subprocess.Popen(p_cmd,
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE)
        (p_out, p_err) = dd_p.communicate()

        if dd_p.returncode != 0:
            raise IOError("Unable to copy %s: %s" % (se_record["path"], p_err))

        dest_se_record = GenericPlugin.getInfo(ss_record, dest_se_record)

        IPlugin.logger.debug("EXIT : copy() - Success")
        return dest_se_record

    @staticmethod   
    def getCurrentClones(ss_record, se_record = None):   
        IPlugin.logger.debug("ENTER: getCurretClones()")   
        IPlugin.logger.debug("EXIT :getCurrentClones() - Success")    
        return []

    @staticmethod   
    def addToAccessGroup(ss_record, access_grp_name, grp_entries):   
        IPlugin.logger.debug("ENTER: addToAccessGroup()")   
        IPlugin.logger.error("addToAccessGroup() is not supported!")   
        raise NoSuchOperationEx(_("addToAccessGroup() is not supported!"))   
 
   
    @staticmethod   
    def removeFromAccessGroup(ss_record, access_grp_name, grp_entries):   
        IPlugin.logger.debug("ENTER: removeAccessGroups()")   
        IPlugin.logger.error("removeAccessGroups() is not supported!")   
        raise NoSuchOperationEx(_("removeAccessGroups() is not supported!"))   
   
    @staticmethod   
    def getAccessGroups(ss_record, se_record = None):   
        IPlugin.logger.debug("ENTER: getAccessGroups()")   
        IPlugin.logger.error("getAccessGroupgs() is not supported!")   
        raise NoSuchOperationEx(_("getAccessGroups() is not supported!"))   
   
   
    @staticmethod   
    def createAccessGroups(ss_record, access_grps):   
        IPlugin.logger.debug("ENTER: createAccessGroups()")   
        IPlugin.logger.error("createAccessGroups() is not supported!")   
        raise NoSuchOperationEx(_("createAccessGroups() is not supported!"))   
   
   
    @staticmethod   
    def removeAccessGroups(ss_record, access_grps):   
        IPlugin.logger.debug("ENTER: removeAccessGroups()")   
        IPlugin.logger.error("removeAccessGroups() is not supported!")   
        raise NoSuchOperationEx(_("removeAccessGroups() is not supported!"))   
   
    @staticmethod   
    def getStorageNames(ss_record):   
        IPlugin.logger.debug("ENTER: getStorageNames()")   
        IPlugin.logger.error("getStorageNames() is not supported!")   
        raise NoSuchOperationEx(_("getStorageNames() is not supported!"))   
   
    @staticmethod   
    def stopPresent(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: stopPresent()")   
        IPlugin.logger.error("stopPresent() is not supported!")   
        raise NoSuchOperationEx(_("stopPresent() is not supported!"))   
   
    @staticmethod   
    def online(ss_record, se_record):   
         IPlugin.logger.debug("ENTER: online()")   
         IPlugin.logger.error("online() is not supported!")   
         raise NoSuchOperationEx(_("online() is not supported!"))   
   
    @staticmethod   
    def offline(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: offline()")   
        IPlugin.logger.error("offline() is not supported!")   
        raise NoSuchOperationEx(_("offline() is not supported!"))   
   
   
    @staticmethod   
    def create(ss_record, se_record, qos = None):   
        IPlugin.logger.debug("ENTER: create()")   
        IPlugin.logger.error("create() is not supported!")   
        raise NoSuchOperationEx(_("create() is not supported!"))   
   
   
    @staticmethod   
    def startPresent(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: startPresent()")   
        IPlugin.logger.error("startPresent() is not supported!")   
        raise NoSuchOperationEx(_("startPresent() is not supported!"))   
   
   
    @staticmethod   
    def resize(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: resize()")   
        IPlugin.logger.error("resize() is not supported!")   
        raise NoSuchOperationEx(_("resize() is not supported!"))   
   
   
    @staticmethod   
    def remove(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: remove()")   
        IPlugin.logger.error("remove() is not supported!")   
        raise NoSuchOperationEx(_("remove() is not supported!"))   
   
    @staticmethod   
    def splitClone(ss_record, se_record):   
        IPlugin.logger.debug("ENTER: splitClone()")   
        IPlugin.logger.error("splitClone() is not supported!")   
        raise NoSuchOperationEx(_("splitClone() is not supported!"))   
   
   
    @staticmethod   
    def cloneFromSnap(ss_record, se_record, snap_se_record, dest_se_record = None, qos = None):   
        IPlugin.logger.debug("ENTER: cloneFromSnap()")   
        IPlugin.logger.error("cloneFromSnap() is not supported!")   
        raise NoSuchOperationEx(_("(cloneFromSnap) is not supported!"))    
   
   
    @staticmethod   
    def getSnapLimits(ss_record, se_record = None):   
        IPlugin.logger.debug("ENTER: getSnapLimits()")   
        IPlugin.logger.error("getSnapLimits() is not supported!")   
        raise NoSuchOperationEx(_("getSnapLimits() is not supported!"))   
   
   
    @staticmethod   
    def isSnapable(ss_record, se_record = None):   
        IPlugin.logger.debug("ENTER: isSnapable()")   
        IPlugin.logger.error("isSnapable() is not supported!")   
        raise NoSuchOperationEx(_("isSnapable() is not supported!"))   
   
   
    @staticmethod   
    def createSnap(ss_record, se_record, snap_se_record = None):   
        IPlugin.logger.debug("ENTER: createSnap()")   
        IPlugin.logger.error("createSnap() is not supported!")   
        raise NoSuchOperationEx(_("createSnap() is not supported!"))   
   
   
    @staticmethod   
    def createMultiSnap(ss_record, se_records, snap_se_record = None):   
        IPlugin.logger.debug("ENTER: createMultiSnap()")   
        IPlugin.logger.error("(createMultiSnap() is not supported!")   
        raise NoSuchOperationEx(_("createMultiSnap(() is not supported!"))   
   
   
    @staticmethod   
    def isRestorable(ss_record, snap_se_record, se_record = None):   
        IPlugin.logger.error("isRestorable() is not supported!")   
        raise NoSuchOperationEx(_("isRestorable() is not supported!"))   
   
   
    @staticmethod   
    def snapRestore(ss_record, snap_se_record, se_record = None):   
        IPlugin.logger.debug("ENTER: snapRestore()")   
        IPlugin.logger.error("snapRestore() is not supported!")   
        raise NoSuchOperationEx(_("snapRestore() is not supported!"))   
   
   
    @staticmethod   
    def snapRemove(ss_record, snap_se_records):   
        IPlugin.logger.debug("ENTER: snapRemove()")   
        IPlugin.logger.error("snapRemove() is not supported!")   
        raise NoSuchOperationEx(_("snapRemove() is not supported!"))   
   
   
    @staticmethod   
    def getCurrentSnaps(ss_record, se_record = None):   
        IPlugin.logger.debug("ENTER: getCurrentSnaps()")   
        IPlugin.logger.error("getCurrentSnaps() is not supported!")   
        raise NoSuchOperationEx(_("getCurrentSnaps() is not supported!"))   
   
   
    @staticmethod   
    def getQoSList(ss_record):   
        IPlugin.logger.debug("ENTER: getQoSList()")   
        IPlugin.logger.error("getQoSList() is not supported!")   
        raise NoSuchOperationEx(_("getQoSList() is not supported!"))   
   
   
    @staticmethod   
    def setQoS(ss_record, se_record, qos):   
        IPlugin.logger.debug("ENTER: setQoS()")   
        IPlugin.logger.error("setQoS() is not supported!")   
        raise NoSuchOperationEx(_("setQoS() is not supported!"))   



    #########################################################################
    ###             START OF THE PLUGIN PRIVATE METHODS                   ###
    #########################################################################


    @staticmethod
    def _get_iscsi_service_status():
        IPlugin.logger.debug("ENTER: _get_iscsi_service_status()")
    
        iscsi_p = subprocess.Popen(["service",
                                  "iscsi",
                                  "status"],
                                  stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE)
        (p_out, p_err) = iscsi_p.communicate()

        if p_out.find('running') < 0:
            return "offline"

        return "online"


    @staticmethod
    def _get_fc_status():
        IPlugin.logger.debug("ENTER: _get_fc_status()")
    
        return "online"


    @staticmethod
    def _get_lun_info_(dev_name):
        lun_info = {}

        lun_info = {"dev_name": dev_name}

        # This is the SG_IO INQUIRY CDB
        get_sg_inq = "12 00 00 00 60 00"
        p_cmd = ("/usr/bin/sg_raw "
                 "--binary "
                 "--nosense "
                 "--request=96 "
                 "--timeout=15 "
                 "%s " +
                 get_sg_inq) % dev_name

        p_cmd = p_cmd.split(" ")
        sgraw_p = subprocess.Popen(p_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE)
        (p_out, p_err) = sgraw_p.communicate()

        if sgraw_p.returncode != 0:
            IPlugin.logger.debug("Unable to issue SG_IO INQ CDB: %s" % p_err)
            raise IOError("Unable to issue SG_IO INQ CDB: %s" % p_err)

        # Use StringIO to read from the string
        str_rdr = StringIO(p_out)
        str_rdr.seek(8)

        vendor  = str_rdr.read(8)
        product = str_rdr.read(16)
        version = str_rdr.read(4)
        serial  = str_rdr.read(32)

        lun_info["vendor"]  = vendor.strip()
        lun_info["product"] = product.strip()
        lun_info["version"] = version.strip()
      

       # This is the SG_IO INQUIRY CDB Page 80
        get_sg_inq = "12 01 80 00 fc 00"
        p_cmd = ("/usr/bin/sg_raw "
                 "--binary "
                 "--nosense "
                 "--request=96 "
                 "--timeout=15 "
                 "%s " +
                 get_sg_inq) % dev_name

        p_cmd = p_cmd.split(" ")
        sgraw_p = subprocess.Popen(p_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE)
        (p_out, p_err) = sgraw_p.communicate()

        if sgraw_p.returncode != 0:
            IPlugin.logger.debug("Unable to issue SG_IO INQ Page 80 CDB to %s: %s" % (dev_name ,p_err))
            raise IOError("Unable to issue SG_IO INQ Page 80 CDB to %s: %s" % (dev_name ,p_err))

        # Use StringIO to read from the string
        str_rdr = StringIO(p_out)
        str_rdr.seek(4)

        serial  = str_rdr.read(32)

        lun_info["serial"]  = serial[:serial.find('\0')].strip()
        
        
        # Use blockdev to get size from the device
        p_cmd = ("/sbin/blockdev "
                 "--getsz "
                 "%s") % dev_name

        p_cmd = p_cmd.split(" ")
        blkdev_p = subprocess.Popen(p_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE)
        (p_out, p_err) = blkdev_p.communicate()

        if blkdev_p.returncode != 0:
            IPlugin.logger.debug("Unable to retrieve blockdev info: %s" % p_err)
            raise IOError("Unable to retrieve blockdev info: %s" % p_err)

        size = 0
        size = int(p_out)
        lun_info["size"] = size * 512


        
        # Get page 0x83: VPD info 12 01 83 00 fc 00
        get_sg_inq = "12 01 83 00 fc 00"
        p_cmd = ("/usr/bin/sg_raw "
                 "--binary "
                 "--nosense "
                 "--request=96 "
                 "--timeout=15 "
                 "%s " +
                 get_sg_inq) % dev_name
        

        p_cmd = p_cmd.split(" ")
        sgraw_p = subprocess.Popen(p_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE)

        (p_out, p_err) = sgraw_p.communicate()

        if sgraw_p.returncode != 0:
            IPlugin.logger.debug("Unable to issue SG_IO INQ Page 83 CDB to %s: %s" % (dev_name ,p_err))
            raise IOError("Unable to issue SG_IO INQ Page 83 CDB to %s: %s" % (dev_name ,p_err))

        str_rdr = StringIO(p_out)
        str_rdr.seek(36)
        page_83  = str_rdr.read(32)

        if page_83:
            lun_info["page83_id"] = page_83.strip()

        # deactivated code for now
        if 0:

            # get paths to the device
            p_cmd = ("/bin/ls "
                     "-l "
                     "%s") % dev_name
    
            p_cmd = p_cmd.split(" ")
    
            path_p = subprocess.Popen(p_cmd,
                                       stdout = subprocess.PIPE,
                                       stderr = subprocess.PIPE)
            (p_out, p_err) = path_p.communicate()
    
            if path_p.returncode != 0:
                raise IOError("Unable to list paths for device %s: %s" % (dev_name, p_err))
    
            index = p_out.find('->') 
            if index < 0:
                return lun_info
    
            parent_dev = p_out[index+6:-1]
    
            lun_info["path"]= []
            lun_info["path"].append("/dev/%s" % parent_dev)
    
            for device in os.listdir("/sys/block/%s/slaves/" % parent_dev):
                lun_info["path"].append("/dev/%s" % device)

        # end of deactivate

        return lun_info


    @staticmethod
    def _get_dev_server(ss_record, mp_dev):
           
        slave = None

        mpl_p = subprocess.Popen(["multipath",
                                  "-l",
                                  os.path.basename(mp_dev)],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE)
        (p_out, p_err) = mpl_p.communicate()

        if mpl_p.returncode != 0:
            IPlugin.logger.error("_get_slave_info() - Unable to obtain "
                                 "multipath info for %s: %s",
                                 mp_dev, p_err)
            raise OperationFailedEx(_("Unable to obtain multipath info for "
                                      "%s: %s" % (mp_dev, p_err)))

        for mpl_line in p_out.splitlines(): 
            mpl_fields = mpl_line.split()
            for index in range(len(mpl_fields)):
                if (mpl_fields[index] == "|-" or mpl_fields[index] == "`-"):
                    slave = mpl_fields[index + 1] #h:b:t:l
                    break

        if slave:  # We only need one slave to get the info
            ls_p = subprocess.Popen(["lsscsi",
                                      "-t",
                                      slave],
                                     stdout = subprocess.PIPE,
                                     stderr = subprocess.PIPE)
            (p_out, p_err) = ls_p.communicate()

            if ls_p.returncode != 0:
                IPlugin.logger.error("Unable to list device %s: %s" %
                                     (mp_dev, p_err))
                raise OperationFailedEx(_("Unable to list device %s: %s" % (mp_dev, p_err)))

            ls_fields = p_out.split()

            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage and ls_fields[2].startswith('iqn.'): 
            
                host_target = ls_fields[2].split(',')
                for target in ss_record["storage_id"]:
                    if target == host_target[0]:
                        return True
                
            if ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
                return True

        return False
